-- Murphi model of Remote Attestation Protocol.
-- We propose a new protocol for remote attestation, by integrating it with SSL
--
-- Protocol Steps
----------------------------------------------------------------------------------
-- (Step 1) C -> S: C, N_c
-- (Step 2) S -> C: N_s, SIG(SK_ca, S, PK_s)
-- (Step 3) C -> S: E(PK_s, Secret, ML), SIG(SK_ca, C, AIK_pub), 
--                  SIK(AIK_pvt, PCR, hash(N_c, N_s, Secret))
-- (Step 4) S -> C: SIG(SK_s, N_c, AIK_pub)
-- (Step 5) C -> S: SIG(AIK_pvt, N_s, PK_s)
--
-- Terminology:
-- C and S are identities of client and server respectively.
-- N_c and N_s are client nonce and server nonce respectively.
-- SK_x represents secret key of X.
-- PK_x represents public key of X.
-- AIK_pub and AIK_pvt are Attestation identity key (pair of RSA public key and
-- private key) generated by TPM on client.
-- E(K, M) is message M encrypted with key K.
-- SIG(K, M) is message M signed by key K.
--
----------------------------------------------------------------------------------

-- Authors:
-- Lavina Jain (jainl@cs.stanford.edu)
-- Jayesh Vyas (jayesh@cs.stanford.edu)

const
  FOUR_STEP:                 true; -- True when in the last step, client waits for server's ack  
  FIVE_STEP:                 true; -- True when server waits for client's ack
  CERT_FORGED:               false; -- True when intruder can forge any signature issued by CA
  SUPER_ATTACKER:            false; -- True when super attacker (can reset PCR/hw attack) is active
  TPM_COMPROMISED:           false; -- TPM on intruder is compromised, intruder knows AIK_private
  SERVER_NONCE_PREDICTABLE:  false; -- true if server's nonce is predictable to the intruder

  NumClients:        1;
  NumServers:        1;
  NumIntruders:      1;
  NetworkSize:       1;

  MaxRandom:     30;

  -- What intruder can remember
  MaxMessages:      20;
  MaxSignatures:    5;
  MaxEncMsgs:       5;
  MaxCertificates:  5;
  MaxNonces:        10;
  MaxSecrets:       5;
  MaxMLs:           5;
  MaxServerSigs:    5;
  MaxClientAcks:    5;
  
  
type
  ClientId:    scalarset (NumClients);
  ServerId:    scalarset (NumServers);
  IntruderId:  scalarset (NumIntruders);
  AgentId:     union {ClientId, ServerId, IntruderId};

  Random:         0..MaxRandom;
  ML:             0..1;

KeyType: enum {
  K_PRIVATE,
  K_PUBLIC
};

Key: record
  ktype:  KeyType;
  val:  Random;
end;

Certificate: record
  owner:  AgentId;
  owner_key:  Key;  -- Public key of the owner
end;

MessageType: enum {
  M_CLIENT_HELLO,
  M_SERVER_HELLO,
  M_CLIENT_KEY_EXCHANGE,
  M_SERVER_ACK,
  M_CLIENT_ACK
};

EncryptedMessage: record
  secret: Random;
  ml: ML;
  encKey: Key; 
end;

-- Hash is implemented as an identity function
Hash: record
  clientNonce:  Random;
  serverNonce:  Random;
  secret:       Random;
end;

-- Represents the data signed by TPM, sent from Client to Server
Signature: record
  pcr:     ML;
  hash:    Hash;
  AIKpub:  Key;
end;

Ack: record
  ownKey:    Key;     -- Private key used for signing
  signedKey: Key;     -- Public key which is being signed (data)
  nonce:     Random;  -- The random nonce sent by the other party
end;

-- A message over the network
Message: record
  source:    AgentId;
  dest:      AgentId;
  mType:     MessageType;
  
  identity:  AgentId;
  nonce:     Random;
  cert:      Certificate;
  encmsg:    EncryptedMessage;
  sign:      Signature;
  --serverSig: ServerSig;
  --clientAck: ClientAck;
  ack:       Ack;
end;

-- Valid fields in the message record for each type of message
-- M_CLIENT_HELLO: identity, nonce (msg 1)
-- M_SERVER_HELLO: nonce, cert (msg 2)
-- M_CLIENT_KEY_EXCHANGE: cert, encmsg, sign (msg 3)
-- M_SERVER_ACK: ServerSig (msg 4)
-- M_CLIENT_ACK: clientAck (msg 5)

ClientState: enum {
  C_SLEEP,
  C_WAIT,
  C_WAIT_ACK,
  C_COMMIT
};

AttackerState: enum {
  A_INACTIVE,
  A_ACTIVE
};

ServerState: enum {
  S_SLEEP,
  S_WAIT,
  S_WAIT_ACK,
  S_ATTESTED,
  S_NOT_ATTESTED
};

Client: record
  state:       ClientState;
  nonce:       Random;
  AIKpub:      Key;
  cert:        Certificate;
  secret:      Random;
  ml:          ML;
  pcr:         ML;

  server:          AgentId;
  serverNonce:     Random;
  serverPublicKey: Key;

  attackerState:   AttackerState;
end;

Server: record
  state:     ServerState;
  nonce:     Random;
  publicKey: Key;
  cert:      Certificate;
  secret:    Random;

  client:          AgentId;
  clientNonce:     Random;
  clientPublicKey: Key;
end;

Intruder: record
  tpmAIK:    Key; -- This is the AIK. Corresponds to the private/public AIK key pair of TPM.
  forgedAIK: Key; -- Intruder uses this key when it can generate fake certificates over this key
  cert:      Certificate;
  secret:    Random;
  ml:        ML;
  pcr:       ML;  -- this will always be 1 because Intruder is always malicious
  nonce:     Random; -- Intruder's random nonce
  sign:      Signature;   -- A saved signature over predictable nonce, taken from a TPM

  messages: multiset[MaxMessages] of Message;
  certificates: multiset[MaxCertificates] of Certificate;
  encMsgs:      multiset[MaxEncMsgs] of EncryptedMessage;
  signatures:   multiset[MaxSignatures] of Signature;
  nonces:       multiset[MaxNonces] of Random;
  secrets:      multiset[MaxSecrets] of Random;
  mls:          multiset[MaxMLs] of ML;
  --serverSigs:   multiset[MaxServerSigs] of ServerSig;
  serverAcks:   multiset[MaxServerSigs] of Ack;
  clientAcks:   multiset[MaxClientAcks] of Ack;
end;

var
  net: multiset[NetworkSize] of Message;
  cli: array[ClientId] of Client;
  ser: array[ServerId] of Server;
  int: array[IntruderId] of Intruder;
  
  nextRandom: Random;
  nextNonce:  Random;
  serNonce:   Random; -- The predictable server nonce
  reusedCliNonce:   Random; -- When server nonce is predictable, intruder and client collude and reuse client Nonce
  reusedCliSecret:  Random; -- When server nonce is predictable, intruder and client collude and reuse client secret

-- Generate a fresh random number

function freshRandom(): Random;
var tmp: Random;
begin
   if (nextRandom = MaxRandom - 1) then -- Last the last randoms is for predictable server nonce
      error "Ran out of random numbers!"
   end;
   tmp := nextRandom;
   nextRandom := nextRandom + 1;
   return tmp;
end;

function keyEqual(k1: Key; k2: Key): boolean;
begin
   return (k1.ktype = k2.ktype) & (k1.val = k2.val);
end;

function certEqual(c1: Certificate; c2: Certificate): boolean;
begin
   return (c1.owner = c2.owner) & keyEqual(c1.owner_key, c2.owner_key)
end;

function encEqual(e1: EncryptedMessage; e2: EncryptedMessage): boolean;
begin
  return (e1.secret = e2.secret) & (e1.ml = e2.ml) & keyEqual(e1.encKey, e2.encKey);
end;

function hashEqual(h1: Hash; h2: Hash): boolean;
begin
  return (h1.clientNonce = h2.clientNonce) & (h1.serverNonce = h2.serverNonce) & (h1.secret = h2.secret);
end;

function signEqual(s1: Signature; s2: Signature): boolean;
begin
  return (s1.pcr = s2.pcr) & hashEqual(s1.hash, s2.hash) & keyEqual(s1.AIKpub, s2. AIKpub);
end;

function msgEqual(m1: Message; m2: Message): boolean;
begin
  return ((m1.source = m2.source) &
          (m1.dest = m2.dest) &
          (m1.mType = m2.mType) &
          (m1.mType = M_CLIENT_HELLO ->
           (m1.identity = m2.identity) & (m1.nonce = m2.nonce)
          ) &
          (m1.mType = M_SERVER_HELLO ->
           (m1.nonce = m2.nonce) & certEqual(m1.cert, m2.cert)
          ) &
          (m1.mType = M_CLIENT_KEY_EXCHANGE ->
           certEqual(m1.cert, m2.cert) & encEqual(m1.encmsg, m2.encmsg) & signEqual(m1.sign, m2.sign)
          )
         );
end;

function AckEqual(c1: Ack; c2: Ack): boolean;
begin
  return keyEqual(c1.ownKey, c2.ownKey) & keyEqual(c1.signedKey, c2.signedKey);
end;

procedure addNonce(non: Random; intId: IntruderId);
begin
    if multisetcount (l:int[intId].nonces,   -- add only if not there
       int[intId].nonces[l] = non) = 0 then
        multisetadd(non, int[intId].nonces);
--        put "adding nonce ";
--        put(non);
--        put "\n";
    end;
--    put "Number of nonces: ";
--    put (multisetcount(l:int[intId].nonces, true));
--    put "\n";
end;

procedure addCert(cert: Certificate; intId: IntruderId);
begin
    if multisetcount (l:int[intId].certificates,   -- add only if not there
       certEqual(int[intId].certificates[l], cert)) = 0 then
        multisetadd(cert, int[intId].certificates);
    end;
end;

procedure addEncmsg(encmsg: EncryptedMessage; intId: IntruderId);
begin
    if( keyEqual(encmsg.encKey, int[intId].tpmAIK) ) then
      if multisetcount(m:int[intId].secrets,
                       encmsg.secret = int[intId].secrets[m]) = 0 then
        multisetadd(encmsg.secret, int[intId].secrets);
      end;
      if multisetcount(n:int[intId].mls,
                       encmsg.ml = int[intId].mls[n]) = 0 then
        multisetadd(encmsg.ml, int[intId].mls);
      end;
      return;
    end;
    if multisetcount (l:int[intId].encMsgs,   -- add only if not there
       encEqual(int[intId].encMsgs[l], encmsg)) = 0 then
        multisetadd(encmsg, int[intId].encMsgs);
    end;
    
end;

procedure addSignature(sign: Signature; intId: IntruderId);
begin
    if multisetcount (l:int[intId].signatures,   -- add only if not there
       signEqual(int[intId].signatures[l], sign)) = 0 then
        multisetadd(sign, int[intId].signatures);
    end;
end;

procedure addServerAck(serAck: Ack; intId: IntruderId);
begin
    if multisetcount (l:int[intId].serverAcks,   -- add only if not there
       AckEqual(int[intId].serverAcks[l], serAck)) = 0 then
        multisetadd(serAck, int[intId].serverAcks);
    end;
end;

procedure addClientAck(cliAck: Ack; intId: IntruderId);
begin
    if multisetcount (l:int[intId].clientAcks,   -- add only if not there
       AckEqual(int[intId].clientAcks[l], cliAck)) = 0 then
        multisetadd(cliAck, int[intId].clientAcks);
    end;
end;

startstate
  nextRandom := 0;
  nextNonce := 0;

  -- Predictable nonces
  serNonce := MaxRandom - 1;

  -- initialize client
  undefine cli;
  for i: ClientId do
    cli[i].state := C_SLEEP;
    cli[i].AIKpub.ktype := K_PUBLIC;
    cli[i].AIKpub.val := freshRandom();
    cli[i].cert.owner := i;
    cli[i].cert.owner_key := cli[i].AIKpub;
    cli[i].server := i;
    
    -- ml and pcr equal to 0 is considered trusted configuration. 1 is considered untrusted configuration
    cli[i].ml := 0;
    cli[i].pcr := 0;
    cli[i].attackerState := A_INACTIVE;
  end;
  
  -- initialize server
  undefine ser;
  for i: ServerId do
    ser[i].state := S_SLEEP;
    ser[i].publicKey.ktype := K_PUBLIC;
    ser[i].publicKey.val := freshRandom();
    ser[i].cert.owner := i;
    ser[i].cert.owner_key := ser[i].publicKey;
    ser[i].client := i;
  end;
  
  -- initialize intruder
  undefine int;
  for i: IntruderId do
    int[i].tpmAIK.ktype := K_PUBLIC;
    int[i].tpmAIK.val := freshRandom();
    int[i].forgedAIK.ktype := K_PUBLIC;
    int[i].forgedAIK.val := freshRandom();
    int[i].cert.owner := i;
    int[i].cert.owner_key := int[i].tpmAIK;
    multisetadd(int[i].cert, int[i].certificates);
    int[i].secret := freshRandom();
    multisetadd(int[i].secret, int[i].secrets);
    int[i].ml := 0;
    -- Adding ml to known MLs
    multisetadd(int[i].ml, int[i].mls);
    -- Adding its own nonce to known nonces
    int[i].nonce := freshRandom();
    multisetadd(int[i].nonce, int[i].nonces);

    -- Adding the predictable nonces to intruder's known nonces
    -- Moreover, intruder can take signature over the predictable nonce
    -- from the client and store it!
    if (SERVER_NONCE_PREDICTABLE) then
      -- Adding the server nonce to known nonces 
      multisetadd(serNonce, int[i].nonces);
      -- Saving the signature generated from client
      int[i].sign.pcr := 0;
      int[i].sign.hash.clientNonce := int[i].nonce;
      int[i].sign.hash.serverNonce := serNonce;     -- The predictable server nonce
      int[i].sign.hash.secret := int[i].secret;
      reusedCliNonce := int[i].nonce;
      reusedCliSecret := int[i].secret;

      -- Taking signature from any of the clients!
      for j: ClientId do
        int[i].sign.AIKpub := cli[j].AIKpub;
      end;
      multisetadd(int[i].sign, int[i].signatures);
    end;
        
    int[i].pcr := 1; -- Since intruder is always corrupt
  end;

  -- initialize network
  undefine net;
end;

--------------------------------------------------------------------------------
-- rules
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- behavior of clients

-- Client i starts protocol with server j (step 1)
-- Client hello message: (C, N_c)
ruleset i: ClientId do
  ruleset j: AgentId do   -- send to server and intruder
    rule 20 "client starts protocol (step 1)"

    !ismember(j,ClientId) &
    cli[i].state = C_SLEEP &
    multisetcount (l:net, true) < NetworkSize

    ==>
    
    var
      outM: Message;   -- outgoing message

    begin
      -- put "client sends hello, src: ";
      -- put (i);
      -- put " dest: ";
      -- put (j);
      -- put "\n";

      -- If server nonce is predictable, then client colludes with intruder, and sends a nonce used earlier
      if (SERVER_NONCE_PREDICTABLE) then
        -- A dirty client not following the protocol => Attacker is active and pcr is having the fingerprint of the attacker
        cli[i].nonce  := reusedCliNonce;
        cli[i].pcr := 1;
        cli[i].attackerState := A_ACTIVE;
      else
        cli[i].nonce  := freshRandom();
      end;

      cli[i].server := j;

      -- Client sends (C, N_c) to server and intruder
      -- Client transits from SLEEP state to WAIT state
      undefine outM;
      outM.source := i;
      outM.dest := j;
      outM.mType := M_CLIENT_HELLO;
      outM.identity := i;
      outM.nonce := cli[i].nonce;
    
      multisetadd(outM, net);
    
      cli[i].state := C_WAIT;

    end;
  end;
end;

-- Client i reacts to nonce received (step 3)
-- Client receives server hello: (N_s, SIG(SK_ca, PK_s))
-- Client sends client key exchange: (encrypted msg, certificate, signature)
-- E(PK_s, secret, ML)||SIG(SK_ca, PK_c)||SIG(SK_c, PCR, HASH(N_c, N_s, secret))
ruleset i: ClientId do
  choose j: net do
    rule 21 "client reacts to nonce received (step 3)"

    cli[i].state = C_WAIT &
    net[j].dest = i --&
--  ismember(net[j].source,IntruderId)
    
    ==>

    var
      outM: Message;   -- outgoing message
      inM:  Message;   -- incoming message

    begin
      inM := net[j];
      multisetremove (j,net);
      -- put "Client received server nonce from ";
      -- put(inM.source);

      if inM.mType = M_SERVER_HELLO then   -- correct message type
        if cli[i].server = inM.cert.owner then  -- correct server identity
          -- save server nonce and public key
          cli[i].serverNonce := inM.nonce;
          cli[i].serverPublicKey := inM.cert.owner_key;
          -- generate sesion secret
          if (SERVER_NONCE_PREDICTABLE) then
            -- A dirty client not following the protocol => Attacker is active and pcr is having the fingerprint of the attacker
            cli[i].secret := reusedCliSecret;
            cli[i].pcr := 1;
            cli[i].attackerState := A_ACTIVE;
          else
            cli[i].secret := freshRandom();
          end;
      
          -- Client sends (encrypted msg, certificate, signature) to server
          -- encrypted msg := E(Server's public key, session secret, measurement list)
          -- certificate := SIG(CA's private key, Client's identity, Client's public key)
          -- signature := SIG(Client's private key, PCR value, HASH(N_c, N_s, session secret))
          undefine outM;
          outM.source := i;
          outM.dest   := cli[i].server;
          outM.mType  := M_CLIENT_KEY_EXCHANGE;
      
          outM.cert := cli[i].cert;

          outM.encmsg.secret := cli[i].secret;
          outM.encmsg.ml     := cli[i].ml;
          outM.encmsg.encKey := cli[i].serverPublicKey;
      
          outM.sign.pcr              := cli[i].pcr;
          outM.sign.hash.clientNonce := cli[i].nonce;
          outM.sign.hash.serverNonce := cli[i].serverNonce;
          outM.sign.hash.secret      := cli[i].secret;
          outM.sign.AIKpub           := cli[i].AIKpub;

          -- put "client sending key exchange\n";
          multisetadd (outM,net);

          -- Client is done if it is a 3-step protocol, and goes to COMMIT state.
          -- else, Client goes to WAIT_ACK state and waits for server ACK.
          if (FIVE_STEP | FOUR_STEP) then
              cli[i].state := C_WAIT_ACK;
          else
              cli[i].state := C_COMMIT;
          end;
        else
          -- put "error: client received hello from wrong server \n";
        end; -- end if
      end; -- end if
    end; -- end begin
  end; -- end choose
end; -- end ruleset

-- Client i reacts to server's signature over nonce (step 5)
-- Client receives server ack: SIG(SK_s, N_c)
-- If five step, Client sends client ack: SIG(SK_c, PK_s)
ruleset i: ClientId do
  choose j: net do
    rule 22 "Client reacts to server's signature over nonce (step 5)"

    cli[i].state = C_WAIT_ACK &
    net[j].dest = i --&
--  ismember(net[j].source,IntruderId)
    
    ==>

    var
      inM:  Message;   -- incoming message
      outM: Message;   -- outgoing message
    begin
      inM := net[j];
      multisetremove (j,net);
      -- put "Client received signed nonce from:";
      -- put(inM.source);

      if inM.mType = M_SERVER_ACK then   -- correct message type
        if keyEqual(cli[i].serverPublicKey, inM.ack.ownKey) then  -- signed by correct server
          -- Ensuring that server signed clients key, it is not talking to intruder
          if (keyEqual(cli[i].AIKpub, inM.ack.signedKey) & cli[i].nonce = inM.ack.nonce) then  -- Signature over correct key and correct nonce
            -- If five step protocol, then client sends back ack(finished message) to server
            if(FIVE_STEP) then
              undefine outM;
              outM.source := i;
              outM.dest := cli[i].server;
              outM.mType := M_CLIENT_ACK;
              --outM.clientAck.ownKey := cli[i].AIKpub;
              --outM.clientAck.signedKey := cli[i].serverPublicKey;
              outM.ack.ownKey := cli[i].AIKpub;
              outM.ack.signedKey := cli[i].serverPublicKey;
              outM.ack.nonce := cli[i].serverNonce;
              multisetadd (outM,net);
            end; -- end if(FIVE_STEP)

            -- Client finishes protocol, goes into COMMIT state
            cli[i].state := C_COMMIT;

          else -- server sent a different N_c
            -- put "error: client received incorrect nonce from correct server\n";
          end; -- end if (nonces)
        else -- server public key does not match
          -- put "error: client received signed nonce from wrong server \n";
        end; -- end if (public key)
      end; -- end if (mType)
    end; -- end begin
  end; -- end choose
end; -- end ruleset

--------------------------------------------------------------------------------
-- behavior of servers

-- Server i reacts to client's hello (step 2)
-- Server receives Client hello: (C, N_c)
-- Server sends Server hello: (N_s, SIG(SK_ca, PK_s))
ruleset i: ServerId do
  choose j: net do
    rule 30 "server reacts to client's hello (step 2)"

    ser[i].state = S_SLEEP &
    net[j].dest = i --&
--  ismember(net[j].source,IntruderId)

    ==>

    var
      outM: Message;   -- outgoing message
      inM:  Message;   -- incoming message

    begin
      inM := net[j];
      multisetremove (j,net);

      -- put "server received client hello from ";
      -- put(inM.source);

      if inM.mType = M_CLIENT_HELLO then   -- correct message type

        -- Store client nonce and client identity
        if (SERVER_NONCE_PREDICTABLE) then
          ser[i].nonce := serNonce;
        else
          ser[i].nonce := freshRandom();
        end;
        ser[i].client := inM.identity;
        ser[i].clientNonce := inM.nonce;
        
        -- Send server hello: (server nonce, server certificate)
        undefine outM;
        outM.source := i;
        outM.dest := inM.identity;
        outM.mType := M_SERVER_HELLO;
        outM.nonce := ser[i].nonce;
        outM.cert := ser[i].cert;

        -- put "sending server hello\n";
        multisetadd (outM,net);
        -- Transit to WAIT state
        ser[i].state := S_WAIT;
      end; -- end if
    end; -- end begin
  end; -- end choose
end; -- end ruleset

-- Server i reacts to client's key exchange (step 4)
-- Server receives client's key exchange: 
-- E(PK_s, secret, ML)||SIG(SK_ca, PK_c)||SIG(AIK, PCR, HASH(N_c, N_s, secret))
ruleset i: ServerId do
  choose j: net do
    rule 31 "server reacts to client's key exchange (step 4)"

    ser[i].state = S_WAIT &
    net[j].dest = i --&
--  ismember(net[j].source,IntruderId)

    ==>

    var
      inM:  Message;   -- incoming message
      outM: Message;   -- outgoing message

    begin
      inM := net[j];
      multisetremove (j,net);
 
      -- put "Server receive client key exchange from ";
      -- put(inM.source);

      if inM.mType = M_CLIENT_KEY_EXCHANGE then   -- correct message type
        if (ser[i].client = inM.cert.owner) then  -- certificate from the client it is talking to
          if (keyEqual(inM.sign.AIKpub, inM.cert.owner_key)) then -- Signed message is signed by the same client
            if (keyEqual(inM.encmsg.encKey, ser[i].publicKey)) then -- Message is encrypted using server's public key
              -- Measurement list matches the signed value in the PCR and is valid (=0)
              if(inM.sign.pcr = inM.encmsg.ml & inM.sign.pcr = 0) then 
                -- match hash value
                if(inM.sign.hash.clientNonce = ser[i].clientNonce &
                   inM.sign.hash.serverNonce = ser[i].nonce &
                   inM.sign.hash.secret = inM.encmsg.secret) then
               
                  -- server stores the session secret received from client
                  ser[i].secret := inM.encmsg.secret;
                  ser[i].clientPublicKey := inM.cert.owner_key;

                  -- If four step or five step protocol, then send server ack
                  -- Send SIG(SK_s, PK_c)
                  if (FIVE_STEP | FOUR_STEP) then
                    undefine outM;
                    outM.source := i;
                    outM.dest   := ser[i].client;
                    outM.mType  := M_SERVER_ACK;
                    outM.ack.signedKey := ser[i].clientPublicKey;
                    outM.ack.ownKey := ser[i].publicKey;
                    outM.ack.nonce := ser[i].clientNonce;
                    multisetadd(outM, net);
                  end;

                  -- Wait for client ack if it is a five step protocol,
                  -- else protocol completes
                  if(FIVE_STEP) then
                    ser[i].state := S_WAIT_ACK;
                   else
                    ser[i].state := S_ATTESTED;
                   end;

                else
                  -- put "Server received invalid hash\n";
                  ser[i].state := S_NOT_ATTESTED;
                end;
              else
                -- put "error: invalid ml/pcr \n";
                ser[i].state := S_NOT_ATTESTED;
              end;
            else
              -- put "error: Server cannot decrypt, message encrypted by wrong key. \n";
              ser[i].state := S_NOT_ATTESTED;
            end;
          else
            -- put "error: forged signature: signing key does not belong to client \n";
            ser[i].state := S_NOT_ATTESTED;
          end;
        else
          -- put "error: server received certificate that does not identify client he is talking to. \n";
          ser[i].state := S_NOT_ATTESTED;
        end;
      end; -- end if(mType)
    end; --end begin
  end; --end choose
end; --end ruleset

-- Server i reacts to client's ack (step 6)
-- Server receives Client ack: SIG(SK_c, PK_s)
-- This rule is fired only in more than 3-step protocol
-- If 4-step protocol, then server finishes protocol, else
-- If 5-step protocol, then server waits for client ack
ruleset i: ServerId do
  choose j: net do
    rule 33 "server reacts to client's ack (step 6)"

    ser[i].state = S_WAIT_ACK &
    net[j].dest = i --&
--      ismember(net[j].source,IntruderId)

    ==>

    var
      inM:  Message;   -- incoming message

    begin
      inM := net[j];
      multisetremove (j,net);
 
      -- put "Server receives client ack from ";
      -- put(inM.source);

      if inM.mType = M_CLIENT_ACK then   -- correct message type
        if (keyEqual(inM.ack.ownKey, ser[i].clientPublicKey)) then -- Signed by correct key
          if(keyEqual(inM.ack.signedKey, ser[i].publicKey) & inM.ack.nonce = ser[i].nonce) then -- Signed correct key and data
            ser[i].state := S_ATTESTED;
          else
            -- put "Wrong key signed in client ack. Client is talking to somebosy else";
            ser[i].state := S_NOT_ATTESTED;
          end;
        else
          -- put "Server received client ack from wrong client\n";
          ser[i].state := S_NOT_ATTESTED;
        end;
      else
        -- put "Wrong message type\n";
      end; --end if(mType)
    end; --end begin
  end; --end choose
end; --end ruleset

--------------------------------------------------------------------------------
-- behavior of intruders

-- intruder i intercepts messages
ruleset i: IntruderId do
  choose j: net do
    rule 10 "intruder intercepts messsages on network"

      !ismember (net[j].source, IntruderId)    -- not for intruders' own messages

    ==>
      var
        msg:  Message;   -- incoming message

      begin

        -- put "intruder intercepts message: ";

        msg := net[j];           -- intercepted message
        multisetremove (j,net);
        -- For each message type store the corresponding parts
        switch msg.mType
        case M_CLIENT_HELLO:
          -- put "Client hello\n";
          addNonce(msg.nonce, i);
          --multisetadd(msg.nonce, int[i].nonces);
          --multisetadd(msg, int[i].messages);

        case M_SERVER_HELLO:
        -- put "server hello\n";
          -- multisetadd(msg.nonce, int[i].nonces);
          -- multisetadd(msg.cert, int[i].certificates);            
          addNonce(msg.nonce, i);
          addCert(msg.cert, i);
          --multisetadd(msg, int[i].messages);
          
        case M_CLIENT_KEY_EXCHANGE:
          -- put "client key exchange\n";
          addCert(msg.cert, i);
          addSignature(msg.sign, i);
          addEncmsg(msg.encmsg, i);
          --multisetadd(msg.cert, int[i].certificates);
          --multisetadd(msg.sign, int[i].signatures);
          --multisetadd(msg.encmsg, int[i].encMsgs);
          --multisetadd(msg, int[i].messages);
        case M_SERVER_ACK:
          -- put "server ack \n";
          addServerAck(msg.ack, i);
        case M_CLIENT_ACK:
          -- put "client ack\n";
          addClientAck(msg.ack, i);
        end;
      --    put "Multiset encMsgs count = ";
      --    put(multisetcount(l:int[i].encMsgs, true));
      --   put "\n";
      end; -- begin
  end; -- choose j: net
end; -- ruleset IntruderId

-- intruder i generates client hello message
ruleset i: IntruderId do            -- arbitrary choice of
  ruleset s: ServerId do            --  source
    ruleset c: AgentId do           --  destination
      choose n: int[i].nonces do    -- known nonces
        rule 91 "Intruder generates client hello"

        !ismember(c, ServerId) & -- Not sending server's identity
        ser[s].state = S_SLEEP & -- generate message only if there is any server waiting for it
        multisetcount (l:net, true) < NetworkSize

        ==>
      
        var
          outM: Message;   -- outgoing message

        begin
          undefine outM;
          outM.source := i;
          outM.dest := s;
          outM.identity := c;
          outM.mType := M_CLIENT_HELLO;
          outM.nonce := int[i].nonces[n];

          -- put "Intruder generates client hello from known nonces, source: ";
          -- put (c);
          -- put " dest: ";
          -- put(s);
          -- put "\n";

          multisetadd (outM,net);
        end; -- begin
      end; -- choose nonces
    end; -- ruleset AgentId
  end; -- ruleset ServerId
end; -- ruleset IntruderID

-- intruder i generates server hello message
ruleset i: IntruderId do                      -- arbitrary choice of
  ruleset c: ClientId do                      -- destination
    choose n: int[i].nonces do                -- known nonces
      choose cert: int[i].certificates do     -- known and stores cerficiates
        rule 11 "Intruder generates server hello"

        cli[c].state = C_WAIT & -- generate message only if there is any client waiting for it
        (multisetcount (l:net, true) < NetworkSize)

        ==>
      
        var
          outM: Message;   -- outgoing message

        begin
          undefine outM;
          outM.source := i;
          outM.dest := c;
          outM.mType := M_SERVER_HELLO;
          outM.nonce := int[i].nonces[n];
          outM.cert := int[i].certificates[cert];

          -- put "Intruder generates server hello from known nonces, dest: ";
          -- put (c);
          -- put "\n";
          -- put "Multiset nonces count = ";
          -- put(multisetcount(l:int[i].nonces, true));
          -- put "\n";

          multisetadd (outM, net);
        end; -- begin
      end; -- choose certificates
    end; -- choose nonces
  end; -- ruleset ClientId
end; -- ruleset IntruderId

-- intruder i generates server hello message by forging server's certificate
ruleset i: IntruderId do             -- arbitrary choice of
  ruleset c: ClientId do             -- destination
    ruleset s: ServerId do           -- server whose certificate is forged
      choose n: int[i].nonces do
        rule 18 "Intruder generates server hello by forging server's certificate"

        CERT_FORGED &
        cli[c].state = C_WAIT & -- generate message only if there is any client waiting for it
        (multisetcount (l:net, true) < NetworkSize)

        ==>
      
        var
          outM: Message;   -- outgoing message

        begin
          undefine outM;
          outM.source := i;
          outM.dest := c;
          outM.mType := M_SERVER_HELLO;
          outM.nonce := int[i].nonces[n];
          outM.cert.owner := s;
          outM.cert.owner_key := int[i].tpmAIK; -- Sending Intruder's public key!

          -- put "Intruder generates server hello from known nonces and forged server's signature, dest: ";
          -- put (c);
          -- put "\n";

          multisetadd (outM, net);
        end; -- begin
      end; -- choose nonces
    end; -- ruleset ServerId
  end; -- ruleset ClientId
end; -- ruleset IntruderId

-- intruder i replays client key exchange message
ruleset i: IntruderId do                    -- arbitrary choice of
  ruleset s: ServerId do                    -- source
    choose en: int[i].encMsgs do            -- saved encrypted messages
      choose cert: int[i].certificates do   -- saved certificates
        choose sig: int[i].signatures do    -- saved signed messages
            
          rule 13 "Intruder replays client key exchange message"

          ser[s].state = S_WAIT & -- generate message only if there is any server waiting for it
          multisetcount (l:net, true) < NetworkSize

          ==>
        
          var
            outM: Message;   -- outgoing message

          begin
            undefine outM;

            outM.source := i;
            outM.dest := s;
            outM.mType := M_CLIENT_KEY_EXCHANGE;
            outM.cert := int[i].certificates[cert];
            outM.sign := int[i].signatures[sig];
            outM.encmsg := int[i].encMsgs[en];

            -- put "Intruder replays client key exchange, dest: ";
            -- put (s);
            -- put "\n";

            multisetadd (outM,net);
          end; -- begin
        end; -- choose signatures
      end; -- choose certificates
    end; -- choose encMsgs
  end; -- ruleset ServerId
end; -- ruleset IntruderId

-- intruder i generates client key exchange message using known secrets
ruleset i: IntruderId do                          -- arbitrary choice of
  choose cert_dst: int[i].certificates do         -- Used for obtaining destination and its public key (for encmsg)
    choose sec: int[i].secrets do                 -- Known secrets
      choose ml: int[i].mls do                    -- Known measurement lists
        choose cert: int[i].certificates do       -- Known certificates
          choose sig: int[i].signatures do        -- Saved signed messages
          
            rule 14 "Intruder generates client key exchange message from known secret"

            ismember(int[i].certificates[cert_dst].owner, ServerId) &
            ser[int[i].certificates[cert_dst].owner].state = S_WAIT & -- generate message only if there is any server waiting for it
            multisetcount (l:net, true) < NetworkSize

            ==>
      
            var
              outM: Message;   -- outgoing message
              s   : ServerId;  -- destination

            begin
              undefine outM;

              s := int[i].certificates[cert_dst].owner;
              -- put "Intruder generates client key exchange using known secret, dest: ";
              -- put (s);
              -- put "\n";

              outM.source := i;
              outM.dest := s;
              outM.mType := M_CLIENT_KEY_EXCHANGE;
              
              -- Replaying the known certificates and signed messages
              outM.cert := int[i].certificates[cert];
              outM.sign := int[i].signatures[sig];

              -- Generating a new encrypted message using known secrets and MLs, and encrypting using public key of destination
              outM.encmsg.secret := int[i].secrets[sec];
              outM.encmsg.ml := int[i].mls[ml];
              outM.encmsg.encKey := int[i].certificates[cert_dst].owner_key;

              multisetadd (outM,net);
            end; -- begin
          end; -- choose signatures
        end; -- choose certificates
      end; -- choose mls
    end; -- choose secrets
  end; -- choose certificates
end; -- ruleset IntruderId

-- intruder i generates client key exchange message using his own signing key, and from known nonces and secrets
-- intruder asks its own TPM to sign the data
ruleset i: IntruderId do                            -- arbitrary choice of
  choose cert_dst: int[i].certificates do           -- Used for obtaining destination and its public key (for encmsg)
    choose n1: int[i].nonces do                     -- Known nonces
      choose n2: int[i].nonces do                   -- Known nonces
        choose sec: int[i].secrets do               -- Learned secrets
          choose ml: int[i].mls do                  -- Known MLs
            choose cert: int[i].certificates do     -- Known certificates
          
              rule 15 "Intruder generates client key exchange message using his own signing key, and from known nonces and secrets"

              ismember(int[i].certificates[cert_dst].owner, ServerId) &
              ser[int[i].certificates[cert_dst].owner].state = S_WAIT & -- generate message only if there is any server waiting for it
              multisetcount (l:net, true) < NetworkSize

              ==>
            
              var
                outM: Message;   -- outgoing message
                s   : ServerId;  -- destination

              begin
                undefine outM;
                s := int[i].certificates[cert_dst].owner;

                -- put "Intruder generates client key exchange with his own signature, dest: ";
                -- put (s);
                -- put "\n";

                outM.source := i;
                outM.dest := s;
                outM.mType := M_CLIENT_KEY_EXCHANGE;
                outM.cert := int[i].certificates[cert]; -- This will include intruder's own certificate also
                
                -- TPM is asked to sign the hash of the following 3 (nonce, nonce and secret)
                outM.sign.hash.clientNonce := int[i].nonces[n1];
                outM.sign.hash.serverNonce := int[i].nonces[n2];
                outM.sign.hash.secret := int[i].secrets[sec];
                -- Unless TPM is compromised, intruder cannot send a signature over valid PCR (pcr = 0 is considered valid). 
                -- Since intruder is a malicious software, so its fingerprint in PCR is dirty (pcr = 1)
                -- Intruder always sends invalid pcr in his own signature, unless he can forge his signature without
                -- using TPM or TPM is compromised
                 
                if (TPM_COMPROMISED) then
                  outM.sign.pcr := 0;
                else
                  outM.sign.pcr := int[i].pcr;
                end;
                outM.sign.AIKpub := int[i].tpmAIK; -- signing with his own key

                outM.encmsg.secret := int[i].secrets[sec];
                outM.encmsg.ml := int[i].mls[ml];
                outM.encmsg.encKey := int[i].certificates[cert_dst].owner_key; -- encryping with receiver's public key

                multisetadd (outM,net);
              end; -- begin
            end; -- choose certificates
          end; -- choose mls
        end; -- choose secrets
      end; -- choose nonces
    end; -- choose nonces
  end; -- choose certificates
end; -- ruleset IntruderId

-- intruder i generates client key exchange message with forged certificate, known nonces and known secrets (all 3 parts of M_CLIENT_KEY_EXCHANGE message are generated)
-- intruder signs the data using a self generated key, and it sends a forged certificate having the new public key
ruleset i: IntruderId do                          -- arbitrary choice of
  choose cert_dst: int[i].certificates do         -- Used for obtaining destination and its public key (for encmsg)
    ruleset c: AgentId do                         -- Client/Intruder whose certificate is forged
      choose n1: int[i].nonces do                 -- Known nonces
        choose n2: int[i].nonces do               -- Known nonces
          choose sec: int[i].secrets do           -- Learned secrets
            choose ml: int[i].mls do              -- Known measurement lists
                  
              rule 19 "Intruder generates client key exchange message with forged certificate, known nonces and known secrets"

              CERT_FORGED &
              !ismember(c, ServerId) &
              ismember(int[i].certificates[cert_dst].owner, ServerId) &
              ser[int[i].certificates[cert_dst].owner].state = S_WAIT & -- generate message only if there is any server waiting for it
              multisetcount (l:net, true) < NetworkSize

              ==>
            
              var
                outM: Message;   -- outgoing message
                s:    ServerId;   -- destination

              begin
                undefine outM;
                s := int[i].certificates[cert_dst].owner;

                -- put "Intruder generates client key exchange with *forged* signature, dest: ";
                -- put (s);
                -- put "\n";

                outM.source := i;
                outM.dest := s;
                outM.mType := M_CLIENT_KEY_EXCHANGE;

                -- Generating a fake certificate
                outM.cert.owner := c;
                outM.cert.owner_key := int[i].forgedAIK;
                
                outM.sign.hash.clientNonce := int[i].nonces[n1];
                outM.sign.hash.serverNonce := int[i].nonces[n2];
                outM.sign.hash.secret := int[i].secrets[sec];
                
                -- Signature is not taken from TPM. Data is signed by the newly generated key.
                -- So intruder can always send a valid PCR and ML.
                outM.sign.pcr := 0;
                outM.sign.AIKpub := int[i].forgedAIK;

                outM.encmsg.secret := int[i].secrets[sec];
                outM.encmsg.ml := int[i].mls[ml];
                outM.encmsg.encKey := int[i].certificates[cert_dst].owner_key;

                multisetadd (outM,net);
              end; -- begin
            end; -- choose ml
          end; -- choose secrets
        end; -- choose nonces
      end; -- choose nonces
    end; -- ruleset ClientId
  end; -- ruleset certificates
end; -- ruleset IntruderId

-- If intruder can predict server nonce, then it generates client key exchange message using the already signed signature data from a valid client, and from known secrets and MLs
ruleset i: IntruderId do                            -- arbitrary choice of
  choose cert_dst: int[i].certificates do           -- Used for obtaining destination and its public key (for encmsg)
      choose sec: int[i].secrets do               -- Learned secrets
        choose ml: int[i].mls do                  -- Known MLs
          choose cert: int[i].certificates do     -- Known certificates
        
            rule 15 "Intruder generates client key exchange message using already signed data over a predictable nonce"

            SERVER_NONCE_PREDICTABLE &
            ismember(int[i].certificates[cert_dst].owner, ServerId) &
            ser[int[i].certificates[cert_dst].owner].state = S_WAIT & -- generate message only if there is any server waiting for it
            multisetcount (l:net, true) < NetworkSize

            ==>
          
            var
              outM: Message;   -- outgoing message
              s   : ServerId;  -- destination

            begin
              undefine outM;
              s := int[i].certificates[cert_dst].owner;

              -- put "Intruder generates client key exchange with sign over predictable nonce, dest: ";
              -- put (s);
              -- put "\n";


              outM.source := i;
              outM.dest := s;
              outM.mType := M_CLIENT_KEY_EXCHANGE;
              outM.cert := int[i].certificates[cert]; -- This will include intruder's own certificate also
              
              outM.sign := int[i].sign;    -- Sending the already taken signature from a valid client

              outM.encmsg.secret := int[i].secrets[sec];
              outM.encmsg.ml := int[i].mls[ml];
              outM.encmsg.encKey := int[i].certificates[cert_dst].owner_key; -- encryping with receiver's public key

              multisetadd (outM,net);
            end; -- begin
          end; -- choose certificates
        end; -- choose mls
      end; -- choose secrets
  end; -- choose certificates
end; -- ruleset IntruderId

-- intruder i replays server ack message
ruleset i: IntruderId do                     -- arbitrary choice of
  ruleset c: ClientId do                     -- destination
    choose serAck: int[i].serverAcks do      -- Saved signed messages
      rule 16 "Intruder replays server ack (signed nonce)"

      cli[c].state = C_WAIT_ACK & -- generate message only if there is any client waiting for it
      multisetcount (l:net, true) < NetworkSize

      ==>
    
      var
        outM: Message;   -- outgoing message

      begin
        undefine outM;
        outM.source := i;
        outM.dest := c;
        outM.mType := M_SERVER_ACK;
        outM.ack := int[i].serverAcks[serAck];

        -- put "Intruder replays signed nonce message, destination: ";
        -- put (c);

        multisetadd (outM,net);
      end; -- begin
    end; -- choose serverAcks
  end; -- ruleset ClientId
end; -- ruleset IntruderId

-- intruder i generates server ack message from known certificates and nonces (signed by TPM)
ruleset i: IntruderId do                -- arbitrary choice of
  choose cert: int[i].certificates do   -- Known certificates
    choose n: int[i].nonces do          -- Known nonces
      rule 17 "Intruder *generates* signed client ack with tpmAIK"

      ismember(int[i].certificates[cert].owner, ClientId) &
      cli[int[i].certificates[cert].owner].state = C_WAIT_ACK & -- generate message only if a client waiting for it
      multisetcount (l:net, true) < NetworkSize

      ==>
    
      var
        outM: Message;   -- outgoing message

      begin
        undefine outM;
        outM.source := i;
        outM.dest := int[i].certificates[cert].owner;
        outM.mType := M_SERVER_ACK;
        outM.ack.ownKey := int[i].tpmAIK;
        outM.ack.signedKey := int[i].certificates[cert].owner_key;
        outM.ack.nonce := int[i].nonces[n];

        -- put "Intruder *generates* client ack, destination: ";
        -- put (outM.dest);
        -- put "\n";

        multisetadd (outM,net);
      end; -- begin
    end; -- choose nonces
  end; -- choose certificates
end; -- ruleset IntruderId

-- intruder i generates server ack message from known certificates and nonces (signed by forged AIK)
ruleset i: IntruderId do                -- arbitrary choice of
  choose cert: int[i].certificates do             -- Known certificates
    choose n: int[i].nonces do          -- Known nonces
      rule 50 "Intruder *generates* client ack with forged AIK"

      CERT_FORGED &
      ismember(int[i].certificates[cert].owner, ClientId) &
      cli[int[i].certificates[cert].owner].state = C_WAIT_ACK & -- generate message only if a client waiting for it
      multisetcount (l:net, true) < NetworkSize

      ==>
    
      var
        outM: Message;   -- outgoing message

      begin
        undefine outM;
        outM.source := i;
        outM.dest := int[i].certificates[cert].owner;
        outM.mType := M_SERVER_ACK;
        outM.ack.ownKey := int[i].forgedAIK;
        outM.ack.signedKey := int[i].certificates[cert].owner_key;
        outM.ack.nonce := int[i].nonces[n];

        -- put "Intruder *generates* client ack (using forgedAIK), destination: ";
        -- put (outM.dest);
        -- put "\n";

        multisetadd (outM,net);
      end; -- begin
    end; -- choose nonces
  end; -- choose cert
end; -- ruleset IntruderId

-- intruder i replays client ack message
ruleset i: IntruderId do                   -- arbitrary choice of
  ruleset s: ServerId do                   -- destination
    choose cliAck: int[i].clientAcks do
      rule 18 "Intruder replays client ack"

      ser[s].state = S_WAIT_ACK & -- generate message only if there is any server waiting for it
      multisetcount (l:net, true) < NetworkSize

      ==>
    
      var
        outM: Message;   -- outgoing message

      begin
        undefine outM;
        outM.source := i;
        outM.dest := s;
        outM.mType := M_CLIENT_ACK;
        outM.ack := int[i].clientAcks[cliAck];

        -- put "Intruder replays client ack, destination: ";
        -- put (s);

        multisetadd (outM,net);
      end; -- begin
    end; -- choose clientAcks
  end; -- ruleset ServerId
end; -- ruleset IntruderId

-- intruder i generates client ack message from known certificates and nonces (signed by TPM)
ruleset i: IntruderId do                          -- arbitrary choice of
  choose cert: int[i].certificates do             -- Known certificates
    choose n: int[i].nonces do                    -- Known nonces
      rule 19 "Intruder *generates* client ack signed by its TPM"

      ismember(int[i].certificates[cert].owner, ServerId) &
      ser[int[i].certificates[cert].owner].state = S_WAIT_ACK & -- generate message only if there is any server waiting for it
      multisetcount (l:net, true) < NetworkSize

      ==>

      var
        outM: Message;   -- outgoing message

      begin
        undefine outM;
        outM.source := i;
        outM.dest := int[i].certificates[cert].owner;
        outM.mType := M_CLIENT_ACK;
        outM.ack.ownKey := int[i].tpmAIK;
        outM.ack.signedKey := int[i].certificates[cert].owner_key;
        outM.ack.nonce := int[i].nonces[n];

        multisetadd (outM,net);
      end; -- begin
    end; -- choose nonces
  end; -- choose certificates
end; -- ruleset IntruderId

-- intruder i generates client ack message from known certificates and nonces (signed by forged AIK)
ruleset i: IntruderId do                          -- arbitrary choice of
  choose cert: int[i].certificates do             -- Known certificates
    choose n: int[i].nonces do                    -- Known nonces
      rule 51 "Intruder *generates* client ack signed by forged AIK"

      CERT_FORGED &
      ismember(int[i].certificates[cert].owner, ServerId) &
      ser[int[i].certificates[cert].owner].state = S_WAIT_ACK & -- generate message only if there is any server waiting for it
      multisetcount (l:net, true) < NetworkSize

      ==>

      var
        outM: Message;   -- outgoing message

      begin
        undefine outM;
        outM.source := i;
        outM.dest := int[i].certificates[cert].owner;
        outM.mType := M_CLIENT_ACK;
        outM.ack.ownKey := int[i].forgedAIK;
        outM.ack.signedKey := int[i].certificates[cert].owner_key;
        outM.ack.nonce := int[i].nonces[n];

        multisetadd (outM,net);
      end; -- begin
    end; -- choose nonces
  end; -- choose certificates
end; -- ruleset IntruderId

--------------------------------------------------------------------------------
-- behavior of attacker on client

-- attacker i is activated at random and its fingerprint if recored in PCR
ruleset i: ClientId do
  rule 40 "Attacker is activated on the client"

  cli[i].state != C_WAIT_ACK &
  cli[i].state != C_COMMIT
  
  ==>

  begin

    cli[i].attackerState := A_ACTIVE;
    -- Attacker can always change ml to represent a valid value
    cli[i].ml := 0;
    -- Only super attacker can change pcr value to a valid value by resetting PCR
    if(SUPER_ATTACKER) then
      cli[i].pcr := 0;
    else
      cli[i].pcr := 1;
    end;
  end; -- begin
end; -- ruleset ClientId

--------------------------------------------------------------------------------
-- Invariants
-- Consistent mutual identification
-- Server Authentication
invariant "If C is talking to S, then S is talking to C"
  forall i: ClientId do
    ismember(cli[i].server, ServerId) &
    cli[i].state = C_COMMIT
    ->
    -- If 3 step, WAIT or ATTESTED or NOT_ATTESTED
    -- If 4 step, ATTESTED or NOT_ATTESTED
    -- If 5 step, WAIT_ACK or ATTESTED or NOT_ATTESTED
    ser[cli[i].server].client = i &
    ((!FOUR_STEP & !FIVE_STEP) ->
    (ser[cli[i].server].state = S_WAIT | 
     ser[cli[i].server].state = S_ATTESTED |
     ser[cli[i].server].state = S_NOT_ATTESTED)) &
    ((FOUR_STEP & !FIVE_STEP) ->
     (ser[cli[i].server].state = S_ATTESTED |
     ser[cli[i].server].state = S_NOT_ATTESTED)) &
    (FIVE_STEP ->
    (ser[cli[i].server].state = S_WAIT_ACK | 
     ser[cli[i].server].state = S_ATTESTED |
     ser[cli[i].server].state = S_NOT_ATTESTED))
        
--    ser[cli[i].server].client = i &
--    (ser[cli[i].server].state = S_WAIT | 
--     ser[cli[i].server].state = S_WAIT_ACK | 
--     ser[cli[i].server].state = S_ATTESTED |
--     ser[cli[i].server].state = S_NOT_ATTESTED)
end;

-- Client Authentication
invariant "If S is talking to C, then C is talking to S"
  forall i: ServerId do
    ismember(ser[i].client, ClientId) &
    ser[i].state = S_ATTESTED
    ->
    -- If 3 step, COMMIT
    -- If 4 step, WAIT_ACK or COMMIT
    -- If 5 step, COMMIT
    cli[ser[i].client].server = i &
    ((!FOUR_STEP & !FIVE_STEP) ->
    cli[ser[i].client].state = C_COMMIT) &
    ((FOUR_STEP & !FIVE_STEP) ->
    (cli[ser[i].client].state = C_WAIT_ACK |
     cli[ser[i].client].state = C_COMMIT)) &
    (FIVE_STEP ->
    cli[ser[i].client].state = C_COMMIT)

--    cli[ser[i].client].server = i &
--    (cli[ser[i].client].state = C_WAIT_ACK |
--    cli[ser[i].client].state = C_COMMIT) 
end;

--
--  Shared secret consistency
--

invariant "If server attests a client, then it shares the same secret"
  forall i: ServerId do
    ismember(ser[i].client, ClientId) &
    ser[i].state = S_ATTESTED &
    cli[ser[i].client].state = C_COMMIT
    ->
    cli[ser[i].client].server = i &
    cli[ser[i].client].secret = ser[i].secret
end;

invariant "If client attests itself to a server, then it shares the same secret with the server"
  forall i: ClientId do
    ismember(cli[i].server, ServerId) &
    cli[i].state = C_COMMIT &
    ser[cli[i].server].state = S_ATTESTED
    ->
    ser[cli[i].server].client = i &
    ser[cli[i].server].secret = cli[i].secret
end;

--
-- Intruder cannot learn any secret
--
invariant "Intruder should not learn server's secret"
  forall s: ServerId do
    forall i: IntruderId do
      ser[s].state = S_ATTESTED &
      ismember(ser[s].client, ClientId) 
    ->
      multisetcount(l:int[i].secrets, ser[s].secret = int[i].secrets[l]) = 0
    end
  end;

invariant "Intruder should not learn client's secret"
  forall c: ClientId do
    forall i: IntruderId do
      cli[c].state = C_COMMIT &
      ismember(cli[c].server, ServerId)
    ->
      multisetcount(l:int[i].secrets, cli[c].secret = int[i].secrets[l]) = 0
    end
  end;

--
-- Invariants related to remote attestation
--
invariant "If server attests a client, then attacker should not be active on that client"
  forall s: ServerId do
    ser[s].state = S_ATTESTED &
    ismember(ser[s].client, ClientId) 
    ->
    cli[ser[s].client].attackerState = A_INACTIVE
  end;

-- Intruder is always malicious
invariant "Server must not attest an intruder" 
  forall s: ServerId do
    ser[s].state = S_ATTESTED
    ->
    !ismember(ser[s].client, IntruderId) 
  end;
